import time
import uuid
import requests
from contextvars import ContextVar
from contextlib import nullcontext

from sovereign import config

_trace_id_ctx_var: ContextVar[str] = ContextVar("trace_id", default="")
_span_id_ctx_var: ContextVar[str] = ContextVar("span_id", default="")


def get_trace_id() -> str:
    return _trace_id_ctx_var.get()


def get_span_id() -> str:
    return _span_id_ctx_var.get()


def generate_128bit():
    return str(uuid.uuid4()).replace("-", "")


def generate_64bit():
    return generate_128bit()[:32]


def timestamp():
    return str(time.time()).replace(".", "")


TRACING = config.tracing
if TRACING is not None:
    TRACING_DISABLED = not TRACING.enabled

    class Tracer:
        def gen_id(self):
            if TRACING.trace_id_128bit:
                trace_id = generate_128bit()
            else:
                trace_id = generate_64bit()
            _trace_id_ctx_var.set(trace_id)
            return trace_id

        def __init__(self, span_name):
            if TRACING_DISABLED:
                return
            span_id = get_span_id()
            self.parent_span_id = None
            if span_id != "":
                # We are already inside a trace context
                self.parent_span_id = span_id
            self.trace_id = get_trace_id()
            self.span_id = self.gen_id()
            self.span_name = span_name

        def __enter__(self):
            if TRACING_DISABLED:
                return nullcontext()
            self.trace = {
                "traceId": self.trace_id,
                "id": self.span_id,
                "name": self.span_name,
                "timestamp": time.time(),
                "tags": TRACING.tags,
            }
            if self.parent_span_id:
                self.trace["parent_span_id"] = self.parent_span_id
            return self

        def __exit__(self, exc_type, exc_value, traceback):
            if TRACING_DISABLED:
                return
            self.trace["duration"] = time.time() - self.trace["timestamp"]
            self.submit()

        def submit(self):
            print(f"{self.span_name}: {self.trace['duration']}")
            try:
                url = f"{TRACING.collector}{TRACING.endpoint}"
                requests.post(url, json=self.trace)
            # pylint: disable=broad-except
            except Exception as e:
                print(f"Failed to submit trace: {self.trace}, Error:{e}")



import os
import sys
from contextvars import ContextVar
from importlib.metadata import version
from typing import Optional

from pydantic import ValidationError
from starlette.templating import Jinja2Templates

from sovereign.context import TemplateContext
from sovereign.logging.bootstrapper import LoggerBootstrapper
from sovereign.schemas import (
    SovereignAsgiConfig,
    SovereignConfig,
    SovereignConfigv2,
    migrate_configs,
    parse_raw_configuration,
)
from sovereign.sources import SourcePoller
from sovereign.statistics import configure_statsd
from sovereign.utils.crypto.crypto import CipherContainer
from sovereign.utils.resources import get_package_file

_request_id_ctx_var: ContextVar[str] = ContextVar("request_id", default="")


def get_request_id() -> str:
    return _request_id_ctx_var.get()


DIST_NAME = "sovereign"

__version__ = version(DIST_NAME)
config_path = os.getenv("SOVEREIGN_CONFIG", "file:///etc/sovereign.yaml")

html_templates = Jinja2Templates(
    directory=str(get_package_file(DIST_NAME, "templates"))
)

if sys.argv[0].endswith("sovereign"):
    migrate_configs()

try:
    config = SovereignConfigv2(**parse_raw_configuration(config_path))
except ValidationError:
    old_config = SovereignConfig(**parse_raw_configuration(config_path))
    config = SovereignConfigv2.from_legacy_config(old_config)
asgi_config = SovereignAsgiConfig()
XDS_TEMPLATES = config.xds_templates()

logs = LoggerBootstrapper(config)
application_logger = logs.application_logger.logger

stats = configure_statsd(config=config.statsd)
poller = None
if config.sources is not None:
    if config.matching is not None:
        matching_enabled = config.matching.enabled
        node_key: Optional[str] = config.matching.node_key
        source_key: Optional[str] = config.matching.source_key
    else:
        matching_enabled = False
        node_key = None
        source_key = None
    poller = SourcePoller(
        sources=config.sources,
        matching_enabled=matching_enabled,
        node_match_key=node_key,
        source_match_key=source_key,
        source_refresh_rate=config.source_config.refresh_rate,
        logger=application_logger,
        stats=stats,
    )

encryption_configs = config.authentication.encryption_configs
server_cipher_container = CipherContainer.from_encryption_configs(
    encryption_configs, logger=application_logger
)

template_context = TemplateContext(
    refresh_rate=config.template_context.refresh_rate,
    refresh_cron=config.template_context.refresh_cron,
    refresh_num_retries=config.template_context.refresh_num_retries,
    refresh_retry_interval_secs=config.template_context.refresh_retry_interval_secs,
    configured_context=config.template_context.context,
    poller=poller,
    encryption_suite=server_cipher_container,
    logger=application_logger,
    stats=stats,
)
if poller is not None:
    poller.lazy_load_modifiers(config.modifiers)
    poller.lazy_load_global_modifiers(config.global_modifiers)



import inspect
from typing import Any, Dict, Optional

from pydantic import BaseModel

from sovereign.utils.entry_point_loader import EntryPointLoader
from sovereign.dynamic_config.loaders import CustomLoader
from sovereign.dynamic_config.deser import ConfigDeserializer


class Loadable(BaseModel):
    path: str
    protocol: str
    serialization: Optional[str] = None

    def load(self, default: Any = None) -> Any:
        if self.protocol not in LOADERS:
            raise KeyError(
                f"Could not find CustomLoader {self.protocol}. Available: {LOADERS}"
            )
        loader_ = LOADERS[self.protocol]

        ser = self.serialization
        if ser is None:
            ser = loader_.default_deser
        if ser not in DESERIALIZERS:
            raise KeyError(
                f"Could not find Deserializer {ser}. Available: {DESERIALIZERS}"
            )
        deserializer = DESERIALIZERS[ser]

        try:
            data = loader_.load(self.path)
            return deserializer.deserialize(data)
        except Exception as original_error:
            if default is not None:
                return default
            raise Exception(
                f"{self.protocol=}, {self.path=}, {self.serialization=}, {original_error=}"
            ) from original_error

    @staticmethod
    def from_legacy_fmt(fmt_string: str) -> "Loadable":
        if "://" not in fmt_string:
            return Loadable(protocol="inline", serialization="string", path=fmt_string)
        try:
            scheme, path = fmt_string.split("://")
        except ValueError:
            raise ValueError(fmt_string)
        try:
            proto, ser = scheme.split("+")
        except ValueError:
            proto, ser = scheme, "yaml"

        if proto in ("python", "module"):
            ser = "raw"
        if proto in ("http", "https"):
            path = "://".join([proto, path])

        return Loadable(
            protocol=proto,
            serialization=ser,
            path=path,
        )


LOADERS: Dict[str, CustomLoader] = {}
for entry_point in EntryPointLoader("loaders").groups["loaders"]:
    custom_loader = entry_point.load()
    func = custom_loader()
    method = getattr(func, "load")
    if not inspect.ismethod(method):
        raise AttributeError(
            f"CustomLoader {entry_point.name} does not implement .load()"
        )
    LOADERS[entry_point.name] = func


DESERIALIZERS: Dict[str, ConfigDeserializer] = {}
for entry_point in EntryPointLoader("deserializers").groups["deserializers"]:
    deserializer = entry_point.load()
    func = deserializer()
    method = getattr(func, "deserialize")
    if not inspect.ismethod(method):
        raise AttributeError(
            f"Deserializer {entry_point.name} does not implement .deserialize()"
        )
    DESERIALIZERS[entry_point.name] = func
